import json

from exspec.models import ExSpec
from exspec.operations import merge_two_dicts
from exspec.utils import TypeUnitValue, ListValue, DimensionFormula


def deflatten_dict(d: dict, name: str) -> ExSpec:
    """
    De-flattens a dictionary, reconstructing complex data types such as ListValue,
    TypeUnitValue, and DimensionFormula from a flattened form. This version also
    collapses keys with common dot-separated prefixes into nested dictionaries.

    Parameters:
    -----------
    - d (dict): The flattened dictionary to deflatten. Keys with specific patterns
      will be converted back into their original data types.
    - name (str): Name of the resulting atomic ExSpec.

    Returns:
    --------
    - ExSpec: The resulting atomic ExSpec, with all flattened structures restored
      and nested according to common key prefixes.
    """
    # First, process the flat dictionary to re-create special types.
    flat_deflattened = _deflatten(d)
    # Then, collapse keys into a nested dictionary.
    nested_deflattened = collapse_common_keys(flat_deflattened)
    return ExSpec(name, [nested_deflattened])


def _deserialize_dimension_formula(dimension_formula_json):
    """
    Converts a serialized dictionary back into a DimensionFormula object.

    Parameters:
    -----------
    - dimension_formula_json (dict): A dictionary representing a serialized
      DimensionFormula.

    Returns:
    --------
    - DimensionFormula: The deserialized DimensionFormula object.
    """
    dims = dimension_formula_json
    items = []
    for dim, exp in dims.items():
        if exp != 0:
            items.append(dim)
            items.append(exp)
    return DimensionFormula(items)


def _deserialize_value(value):
    """
    Deserializes a value, handling both atomic values and complex nested
    structures (lists, dictionaries).

    Parameters:
    -----------
    - value: The value to deserialize, which may be a string, JSON object,
      or serialized structure.

    Returns:
    --------
    - The deserialized value, either as an atomic value or a complex structure.
    """
    try:
        # Attempt to parse value as JSON
        loaded_value = json.loads(value)

        # Recursively deserialize lists and dictionaries
        if isinstance(loaded_value, list):
            return [_deserialize_value(item) for item in loaded_value]
        elif isinstance(loaded_value, dict):
            # Handle DimensionFormula deserialization
            if "__DimensionFormula__" in loaded_value:
                return _deserialize_dimension_formula(loaded_value["__DimensionFormula__"])
            # Recursively deserialize nested dictionaries
            return {k: _deserialize_value(v) for k, v in loaded_value.items()}
        elif value in ('str', 'int', 'float'):
            type_map = {
                "int": int,
                "float": float,
                "str": str
            }
            return type_map.get(value)
        else:
            return loaded_value  # Base case: Return atomic values
    except (TypeError, ValueError):
        # Return the original value if it cannot be parsed as JSON
        return value


def _deflatten(sub_d):
    """
    A helper function to deflatten a dictionary by reconstructing complex
    types such as TypeUnitValue and ListValue. This function operates on the
    flat dictionary (with dot-separated keys) and returns a new dictionary
    where the keys are still flat.

    Parameters:
    -----------
    - sub_d (dict): The flattened dictionary to deflatten.

    Returns:
    --------
    - dict: The deflattened dictionary with keys still in their flattened form.
    """
    deflattened_dict = {}
    for k, v in list(sub_d.items()):
        if k in sub_d.keys():
            if '._TypeUnitValue.' in k:
                # Convert back to TypeUnitValue
                base_key = k.rsplit('._TypeUnitValue._value', 1)[0]
                if sub_d.get(f"{base_key}._TypeUnitValue._value._ListValue"):
                    value = ListValue(items=json.loads(sub_d.pop(f"{base_key}._TypeUnitValue._value._ListValue")))
                else:
                    value = sub_d.pop(f"{base_key}._TypeUnitValue._value")
                unit = sub_d.pop(f"{base_key}._TypeUnitValue._unit", None)
                prefix = sub_d.pop(f"{base_key}._TypeUnitValue._prefix", None)
                prop_type = None
                if sub_d.get(f"{base_key}._TypeUnitValue._type._DimensionFormula"):
                    prop_type = _deserialize_dimension_formula(
                        sub_d.pop(f"{base_key}._TypeUnitValue._type._DimensionFormula"))
                elif sub_d.get(f"{base_key}._TypeUnitValue._type"):
                    prop_type = _deserialize_value(sub_d.pop(f"{base_key}._TypeUnitValue._type"))
                deflattened_dict = merge_two_dicts(deflattened_dict,
                                                   {base_key: TypeUnitValue(value, prefix, unit, prop_type)})
            elif '._ListValue' in k:
                # Convert back to ListValue
                original_key = k.replace('._ListValue', '')
                deflattened_dict = merge_two_dicts(deflattened_dict,
                                                   {original_key: ListValue(items=_deserialize_value(v))})
                sub_d.pop(k)
            else:
                deflattened_dict = merge_two_dicts(deflattened_dict, {k: v})
                sub_d.pop(k)
    return deflattened_dict


def collapse_common_keys(flat_d: dict) -> dict:
    """
    Collapses a flattened dictionary with dot-separated keys into a nested
    dictionary. For example, a key like:

      "workflow.atomic.record.resolution"

    will be collapsed into:

      {
        "workflow": {
          "atomic": {
            "record": {
              "resolution": <value>
            }
          }
        }
      }

    Parameters:
    -----------
    - flat_d (dict): The dictionary with flattened keys.

    Returns:
    --------
    - dict: A nested dictionary with the common key hierarchies collapsed.
    """
    nested = {}
    for key, value in flat_d.items():
        parts = key.split('.')
        d = nested
        for part in parts[:-1]:
            if part not in d:
                d[part] = {}
            d = d[part]
        d[parts[-1]] = value
    return nested
